"""
This module implements discrete transforms for real signals:

- [Discrete Cosine Transform](https://w.wiki/AQEt)
- [Discrete Sine Transform](https://w.wiki/ATnn)

The implementations relies on the FFT under-the-hood, with memory-saving
tricks (borrowed from [`cupy`](https://github.com/cupy/cupy)).

The table belows lists all functions implemented in the module.
In addition to these, wrappers of the form
`#!python FUNCTYPE = partial(FUNC, type=TYPE)` are defined. For example:
```python
dct1 = partial(dct, type=1)
idct1 = partial(idct, type=1)
dctn1 = partial(dctn, type=1)
```

Functions
---------
dct
    One-dimensional Discrete Cosine Transform (DCT)
dst
    One-dimensional Discrete Sine Transform (DST)
idct
    One-dimensional Inverse Discrete Cosine Transform (DCT)
idst
    One-dimensional Inverse Discrete Sine Transform (DST)
dctn
    N-dimensional Discrete Cosine Transform (IDCT)
dstn
    N-dimensional Discrete Sine Transform (IDST)
idctn
    N-dimensional Inverse Discrete Cosine Transform (IDCT)
idstn
    N-dimensional Inverse Discrete Sine Transform (IDST)
```

"""
__all__ = [
    # 1D
    'dct', 'dst', 'idct', 'idst',
    # ND
    'dctn', 'dstn', 'idctn', 'idstn',
    # convenience wrappers
    'dct1', 'dct2', 'dct3', 'dct4',
    'dst1', 'dst2', 'dst3', 'dst4',
    'idct1', 'idct2', 'idct3', 'idct4',
    'idst1', 'idst2', 'idst3', 'idst4',
    'dctn1', 'dctn2', 'dctn3', 'dctn4',
    'dstn1', 'dstn2', 'dstn3', 'dstn4',
    'idctn1', 'idctn2', 'idctn3', 'idctn4',
    'idstn1', 'idstn2', 'idstn3', 'idstn4',
]
from torch import Tensor
from typing import Optional
from functools import partial
from ._realtransforms_autograd import DCTN, DSTN, flipnorm, fliptype
_IMPLEMENTED_TYPES = (1, 2, 3)


def dct(
    x: Tensor,
    dim: int = -1,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return the Discrete Cosine Transform

    !!! warning "Type IV not implemented"

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : int
        Dimensions over which the DCT is computed.
        Default is the last one.
    norm : {"backward", "ortho", "forward"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = -1
    if type in _IMPLEMENTED_TYPES:
        return DCTN.apply(x, type, dim, norm)
    else:
        raise ValueError('DCT only implemented for types I-IV')


def idct(
    x: Tensor,
    dim: int = -1,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return the Inverse Discrete Cosine Transform

    !!! warning
        Type IV not implemented

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : int
        Dimensions over which the DCT is computed.
        Default is the last one.
    norm : {"backward", "ortho", "forward"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = -1
    norm = flipnorm[norm or "backward"]
    type = fliptype[type]
    return dct(x, dim, norm, type)


def dst(
    x: Tensor,
    dim: int = -1,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return the Discrete Sine Transform

    !!! warning "Type IV not implemented"

    !!! warning
        `dst(..., norm="ortho")` yields a different result than `scipy`
        and `cupy` for types 2 and 3. This is because their DST is not
        properly orthogonalized. Use `norm="ortho_scipy"` to get results
        matching their implementation.

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : int
        Dimensions over which the DCT is computed.
        Default is the last one.
    norm : {"backward", "ortho", "forward", "ortho_scipy"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = -1
    if type in _IMPLEMENTED_TYPES:
        return DSTN.apply(x, type, dim, norm)
    else:
        raise ValueError('DST only implemented for types I-IV')


def idst(
    x: Tensor,
    dim: int = -1,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return the Inverse Discrete Sine Transform

    !!! warning "Type IV not implemented"

    !!! warning
        `idst(..., norm="ortho")` yields a different result than `scipy`
        and `cupy` for types 2 and 3. This is because their DST is not
        properly orthogonalized. Use `norm="ortho_scipy"` to get results
        matching their implementation.

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : int
        Dimensions over which the DCT is computed.
        Default is the last one.
    norm : {"backward", "ortho", "forward", "ortho_scipy"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = -1
    norm = flipnorm[norm or "backward"]
    type = fliptype[type]
    return dst(x, dim, norm, type)


def dctn(
    x: Tensor,
    dim: Optional[int] = None,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return multidimensional Discrete Cosine Transform
    along the specified axes.

    !!! warning "Type IV not implemented"

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : [sequence of] int
        Dimensions over which the DCT is computed.
        If not given, all dimensions are used.
    norm : {"backward", "ortho", "forward"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = list(range(x.ndim))
    if type in _IMPLEMENTED_TYPES:
        return DCTN.apply(x, type, dim, norm)
    else:
        raise ValueError('DCT only implemented for types I-IV')


def idctn(
    x: Tensor,
    dim: Optional[int] = None,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return multidimensional Inverse Discrete Cosine Transform
    along the specified axes.

    !!! warning "Type IV not implemented"

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : [sequence of] int
        Dimensions over which the DCT is computed.
        If not given, all dimensions are used.
    norm : {"backward", "ortho", "forward"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = list(range(x.ndim))
    norm = flipnorm[norm or "backward"]
    type = fliptype[type]
    return dctn(x, dim, norm, type)


def dstn(
    x: Tensor,
    dim: Optional[int] = None,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return multidimensional Discrete Sine Transform
    along the specified axes.

    !!! warning "Type IV not implemented"

    !!! warning
        `dstn(..., norm="ortho")` yields a different result than `scipy`
        and `cupy` for types 2 and 3. This is because their DST is not
        properly orthogonalized. Use `norm="ortho_scipy"` to get results
        matching their implementation.

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : [sequence of] int
        Dimensions over which the DCT is computed.
        If not given, all dimensions are used.
    norm : {"backward", "ortho", "forward", "ortho_scipy"}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = list(range(x.ndim))
    if type in _IMPLEMENTED_TYPES:
        return DSTN.apply(x, type, dim, norm)
    else:
        raise ValueError('DST only implemented for types I-IV')


def idstn(
    x: Tensor,
    dim: Optional[int] = None,
    norm: str = 'backward',
    type: int = 2,
) -> Tensor:
    """Return multidimensional Inverse Discrete Sine Transform
    along the specified axes.

    !!! warning "Type IV not implemented"

    !!! warning
        `idstn(..., norm="ortho")` yields a different result than `scipy`
        and `cupy` for types 2 and 3. This is because their DST is not
        properly orthogonalized. Use `norm="ortho_scipy"` to get results
        matching their implementation.

    Parameters
    ----------
    x : tensor
        The input tensor
    dim : [sequence of] int
        Dimensions over which the DCT is computed.
        If not given, all dimensions are used.
    norm : {"backward", "ortho", "forward", "ortho_scipy}
        Normalization mode. Default is "backward".
    type: {1, 2, 3, 4}
        Type of the DCT. Default type is 2.

    Returns
    -------
    y : tensor
        The transformed tensor.

    """
    if dim is None:
        dim = list(range(x.ndim))
    norm = flipnorm[norm or "backward"]
    type = fliptype[type]
    return dstn(x, dim, norm, type)


dct1 = partial(dct, type=1)
dct2 = partial(dct, type=2)
dct3 = partial(dct, type=3)
dct4 = partial(dct, type=4)

idct1 = partial(idct, type=1)
idct2 = partial(idct, type=2)
idct3 = partial(idct, type=3)
idct4 = partial(idct, type=4)

dst1 = partial(dst, type=1)
dst2 = partial(dst, type=2)
dst3 = partial(dst, type=3)
dst4 = partial(dst, type=4)

idst1 = partial(idst, type=1)
idst2 = partial(idst, type=2)
idst3 = partial(idst, type=3)
idst4 = partial(idst, type=4)

dctn1 = partial(dctn, type=1)
dctn2 = partial(dctn, type=2)
dctn3 = partial(dctn, type=3)
dctn4 = partial(dctn, type=4)

idctn1 = partial(idctn, type=1)
idctn2 = partial(idctn, type=2)
idctn3 = partial(idctn, type=3)
idctn4 = partial(idctn, type=4)

dstn1 = partial(dstn, type=1)
dstn2 = partial(dstn, type=2)
dstn3 = partial(dstn, type=3)
dstn4 = partial(dstn, type=4)

idstn1 = partial(idstn, type=1)
idstn2 = partial(idstn, type=2)
idstn3 = partial(idstn, type=3)
idstn4 = partial(idstn, type=4)



import torch
from . import _realtransforms_from_fft as F


flipnorm = {
    'forward': 'backward',
    'backward': 'forward',
    'ortho': 'ortho',
    'ortho_scipy': 'ortho_scipy',
}
fliptype = {1: 1, 2: 3, 3: 2, 4: 4}


class DCTN(torch.autograd.Function):

    @staticmethod
    def forward(ctx, x, type, dim, norm):
        norm = norm or 'backward'
        if type not in (2, 3) and norm == 'ortho_scipy':
            norm = 'ortho'
        ctx.dim = dim
        ctx.norm = norm
        ctx.type = type
        return F.dctn(x, type=type, dim=dim, norm=norm)

    @staticmethod
    def backward(ctx, x):
        if ctx.norm == 'ortho':
            # orthogonal: inverse == transpose
            x = F.idctn(x, type=ctx.type, dim=ctx.dim, norm=ctx.norm)
            return x, None, None, None

        scipy = 'scipy' in ctx.norm
        norm = "backward" if scipy else ctx.norm
        type = fliptype[ctx.type]
        prm = dict(dim=ctx.dim, type=type, square=True)
        spprm = dict(dim=ctx.dim, type=type, scipy=True, inverse=True)
        if type in (1, 3):
            x = x.clone()
        if scipy:
            x = F.ortho_prescale_(x, **spprm)
        x = F.ortho_prescale_(x, **prm)
        x = F.dctn(x, type=type, dim=ctx.dim, norm=norm)
        x = F.ortho_postscale_(x, **prm)
        if scipy:
            x = F.ortho_postscale_(x, **spprm)
            x = F.ortho_scale_(x, ctx.dim, type=type)
        return x, None, None, None


class DSTN(torch.autograd.Function):

    @staticmethod
    def forward(ctx, x, type, dim, norm):
        norm = norm or 'backward'
        if type not in (2, 3) and norm == 'ortho_scipy':
            norm = 'ortho'
        ctx.dim = dim
        ctx.norm = norm
        ctx.type = type
        return F.dstn(x, type=type, dim=dim, norm=norm)

    @staticmethod
    def backward(ctx, x):
        if ctx.norm == 'ortho':
            # orthogonal: inverse == transpose
            x = F.idstn(x, type=ctx.type, dim=ctx.dim, norm=ctx.norm)
            return x, None, None, None

        scipy = 'scipy' in ctx.norm
        norm = "backward" if scipy else ctx.norm
        type = fliptype[ctx.type]
        prm = dict(
            dim=ctx.dim, type=type, square=True, dst=True,
        )
        spprm = dict(
            dim=ctx.dim, type=type, scipy=True, inverse=True, dst=True,
        )
        if type in (1, 3):
            x = x.clone()
        if scipy:
            x = F.ortho_prescale_(x, **spprm)
        x = F.ortho_prescale_(x, **prm)
        x = F.dstn(x, type=type, dim=ctx.dim, norm=norm)
        x = F.ortho_postscale_(x, **prm)
        if scipy:
            x = F.ortho_postscale_(x, **spprm)
            x = F.ortho_scale_(x, ctx.dim, type=type)
        return x, None, None, None



"""
This module contains functions that wrap out-of-bound indices back in-bounds,
according to some boundary condition.

Functions
---------
replicate
    Apply replicate boundary conditions to an index.
    **Aliases:**
    [`border`][bounds.indexing.border],
    [`nearest`][bounds.indexing.nearest],
    [`repeat`][bounds.indexing.repeat],
    [`edge`][bounds.indexing.edge].
dft
    Apply DFT boundary conditions to an index.
    **Aliases:**
    [`wrap`][bounds.indexing.wrap],
    [`gridwrap`][bounds.indexing.gridwrap],
    [`circular`][bounds.indexing.circular],
    [`circulant`][bounds.indexing.circulant].
dct1
    Apply DCT-I boundary conditions to an index.
    **Alias:** [`mirror`][bounds.indexing.mirror]
dct2
    Apply DCT-II boundary conditions to an index.
    **Aliases**:
    [`reflect`][bounds.indexing.reflect],
    [`gridmirror`][bounds.indexing.gridmirror],
    [`neumann`][bounds.indexing.neumann].
dst1
    Apply DST-I boundary conditions to an index.
    **Alias:** [`antimirror`][bounds.indexing.antimirror]
dst2
    Apply DST-II boundary conditions to an index.
    **Aliases**:
    [`antireflect`][bounds.indexing.antireflect],
    [`dirichlet`][bounds.indexing.dirichlet].
nocheck
    Do not wrap indices (assume they are inbounds)

"""
__all__ = [
    'nocheck',
    'replicate', 'repeat', 'nearest', 'border', 'edge',
    'dft', 'wrap', 'gridwrap', 'circular', 'circulant',
    'dct1', 'mirror',
    'dct2', 'reflect', 'reflection', 'gridmirror', 'neumann',
    'dst1', 'antimirror',
    'dst2', 'antireflect', 'dirichlet',
]
import torch
from torch import Tensor
from typing import Tuple
from ._utils import floor_div_int


def nocheck(i, n):
    """Assume all indices are inbounds"""
    return i, 1


def replicate(i, n):
    """Apply replicate (nearest/border) boundary conditions to an index

    !!! info "Aliases"
        [`border`][bounds.indexing.border],
        [`nearest`][bounds.indexing.nearest],
        [`repeat`][bounds.indexing.repeat]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : {1, 0, -1}
        Sign of the transformation (always 1 for replicate)

    """
    return (replicate_script(i, n) if torch.is_tensor(i) else
            replicate_int(i, n))


def replicate_int(i, n):
    return min(max(i, 0), n-1), 1


@torch.jit.script
def replicate_script(i, n: int) -> Tuple[Tensor, int]:
    return i.clamp(min=0, max=n-1), 1


def dft(i, n):
    """Apply DFT (circulant/wrap) boundary conditions to an index

    !!! info "Aliases"
        [`wrap`][bounds.indexing.wrap],
        [`circular`][bounds.indexing.circular]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : {1, 0, -1}
        Sign of the transformation (always 1 for dft)

    """
    return dft_script(i, n) if torch.is_tensor(i) else dft_int(i, n)


def dft_int(i, n):
    return i % n, 1


@torch.jit.script
def dft_script(i, n: int) -> Tuple[Tensor, int]:
    return i.remainder(n), 1


def dct2(i, n):
    """Apply DCT-II (reflect) boundary conditions to an index

    !!! info "Aliases"
        [`reflect`][bounds.indexing.reflect],
        [`neumann`][bounds.indexing.neumann]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : {1, 0, -1}
        Sign of the transformation (always 1 for dct2)

    """
    return dct2_script(i, n) if torch.is_tensor(i) else dct2_int(i, n)


def dct2_int(i: int, n: int) -> Tuple[int, int]:
    n2 = n * 2
    i = (n2 - 1) - i if i < 0 else i
    i = i % n2
    i = (n2 - 1) - i if i >= n else i
    return i, 1


@torch.jit.script
def dct2_script(i, n: int) -> Tuple[Tensor, int]:
    n2 = n * 2
    i = torch.where(i < 0, (n2 - 1) - i, i)
    i = i.remainder(n2)
    i = torch.where(i >= n, (n2 - 1) - i, i)
    return i, 1


def dct1(i, n):
    """Apply DCT-I (mirror) boundary conditions to an index

    !!! info "Aliases"
        [`mirror`][bounds.indexing.mirror]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : {1, 0, -1}
        Sign of the transformation (always 1 for dct1)

    """
    return dct1_script(i, n) if torch.is_tensor(i) else dct1_int(i, n)


def dct1_int(i: int, n: int) -> Tuple[int, int]:
    if n == 1:
        return 0, 1
    n2 = (n - 1) * 2
    i = abs(i) % n2
    i = n2 - i if i >= n else i
    return i, 1


@torch.jit.script
def dct1_script(i, n: int) -> Tuple[Tensor, int]:
    if n == 1:
        return torch.zeros_like(i), 1
    n2 = (n - 1) * 2
    i = i.abs().remainder(n2)
    i = torch.where(i >= n, n2 - i, i)
    return i, 1


def dst1(i, n):
    """Apply DST-I (antimirror) boundary conditions to an index

    !!! info "Aliases"
        [`antimirror`][bounds.indexing.antimirror]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : [tensor of] {1, 0, -1}
        Sign of the transformation

    """
    return dst1_script(i, n) if torch.is_tensor(i) else dst1_int(i, n)


def dst1_int(i: int, n: int) -> Tuple[int, int]:
    n2 = 2 * (n + 1)

    # sign
    ii = (2*n - i) if i < 0 else i
    ii = (ii % n2) % (n + 1)
    x = 0 if ii == n else 1
    x = -x if (i / (n + 1)) % 2 >= 1 else x

    # index
    i = -i - 2 if i < 0 else i
    i = i % n2
    i = (n2 - 2) - i if i > n else i
    i = min(max(i, 0), n-1)
    return i, x


@torch.jit.script
def dst1_script(i, n: int) -> Tuple[Tensor, Tensor]:
    n2 = 2 * (n + 1)

    # sign
    #   zeros
    ii = torch.where(i < 0, 2*n - i, i).remainder(n2).remainder(n + 1)
    x = (ii != n).to(torch.int8)
    #   +/- ones
    ii = torch.where(i < 0, n - 1 - i, i)
    x = torch.where(floor_div_int(ii, n + 1).remainder(2) >= 1, -x, x)

    # index
    i = torch.where(i < 0, -2 - i, i)
    i = i.remainder(n2)
    i = torch.where(i > n, (n2 - 2) - i, i)
    i = i.clamp(0, n-1)
    return i, x


def dst2(i, n):
    """Apply DST-II (antireflect) boundary conditions to an index

    !!! info "Aliases"
        [`antireflect`][bounds.indexing.antireflect],
        [`dirichlet`][bounds.indexing.dirichlet]

    Parameters
    ----------
    i : int or tensor
        Index
    n : int
        Length of the field of view

    Returns
    -------
    i : int or tensor
        Index that falls inside the field of view [0, n-1]
    s : [tensor of] {1, 0, -1}
        Sign of the transformation (always 1 for dct1)

    """
    return dst2_script(i, n) if torch.is_tensor(i) else dst2_int(i, n)


def dst2_int(i: int, n: int) -> Tuple[int, int]:
    x = -1 if (i/n) % 2 >= 1 else 1
    return dct2_int(i, n)[0], x


@torch.jit.script
def dst2_script(i, n: int) -> Tuple[Tensor, Tensor]:
    x = torch.ones([1], dtype=torch.int8, device=i.device)
    x = torch.where(floor_div_int(i, n).remainder(2) >= 1, -x, x)
    return dct2_script(i, n)[0], x


nearest = border = repeat = edge = replicate
"""Alias for [`replicate`][bounds.indexing.replicate]"""

reflect = reflection = gridmirror = neumann = dct2
"""Alias for [`dct2`][bounds.indexing.dct2]"""

mirror = dct1
"""Alias for [`dct1`][bounds.indexing.dct1]"""

antireflect = dirichlet = dst2
"""Alias for [`dst2`][bounds.indexing.dst2]"""

antimirror = dst1
"""Alias for [`dst1`][bounds.indexing.dst1]"""

wrap = gridwrap = circular = circulant = circulant = dft
"""Alias for [`dft`][bounds.indexing.dft]"""



"""
This module reimplements [`torch.nn.functional.pad`][] and [`torch.roll`][]
with a larger set of boundary conditions.

Functions
---------
pad
    Pad a tensor
roll
    Roll a tensor
ensure_shape
    Pad/crop a tensor so that it has a given shape
"""
__all__ = ['pad', 'roll', 'ensure_shape']
import torch
import math
from torch import Tensor
from typing import Optional
from numbers import Number
from . import indexing
from .types import to_fourier, BoundLike, SequenceOrScalar
from ._utils import prod, ensure_list, meshgrid_list_ij, sub2ind_list


def pad(
    inp: Tensor,
    padsize: SequenceOrScalar[int],
    mode: SequenceOrScalar[BoundLike] = 'constant',
    value: Number = 0,
    side: Optional[str] = None
):
    """Pad a tensor.

    This function is a bit more generic than torch's native pad
    (`torch.nn.functional.pad`), but probably a bit slower:

    - works with any input type
    - works with arbitrarily large padding size
    - crops the tensor for negative padding values
    - implements additional padding modes

    When used with defaults parameters (`side=None`), it behaves
    exactly like `torch.nn.functional.pad`

    !!! info "Boundary modes"
        Like in PyTorch's `pad`, boundary modes include:

        - `'circular'`  (or `'dft'`)
        - `'mirror'`    (or `'dct1'`)
        - `'reflect'`   (or `'dct2'`)
        - `'replicate'` (or `'nearest'`)
        - `'constant'`  (or `'zero'`)

        as well as the following new modes:

        - `'antimirror'`    (or `'dst1'`)
        - `'antireflect'`   (or `'dst2'`)

    !!! info "Side modes"
        Side modes are `'pre'` (or `'left'`), `'post'` (or `'right'`),
        `'both'` or `None`.

        - If side is not `None`, `inp.dim()` values (or less) should be
          provided.
        - If side is `None`, twice as many values should be provided,
          indicating different padding sizes for the `'pre'` and `'post'`
          sides.
        - If the number of padding values is less than the dimension of the
          input tensor, zeros are prepended.

    Parameters
    ----------
    inp : tensor
        Input tensor
    padsize : SequenceOrScalar[int]
        Amount of padding in each dimension.
    mode : SequenceOrScalar[BoundLike]
        Padding mode
    value : scalar
        Value to pad with in mode `'constant'`.
    side : "{'pre', 'post', 'both', None}"
        Use padsize to pad on left side (`'pre'`), right side (`'post'`) or
        both sides (`'both'`). If `None`, the padding side for the left and
        right sides should be provided in alternate order.

    Returns
    -------
    out : tensor
        Padded tensor.

    """
    # Argument checking
    mode = to_fourier(mode)
    mode = ensure_list(mode, len(padsize) // (1 if side else 2))

    padsize = tuple(padsize)
    if not side:
        if len(padsize) % 2:
            raise ValueError('Padding length must be divisible by 2')
        padpre = padsize[::2]
        padpost = padsize[1::2]
    else:
        side = side.lower()
        if side == 'both':
            padpre = padsize
            padpost = padsize
        elif side in ('pre', 'left'):
            padpre = padsize
            padpost = (0,) * len(padpre)
        elif side in ('post', 'right'):
            padpost = padsize
            padpre = (0,) * len(padpost)
        else:
            raise ValueError(f'Unknown side `{side}`')
    padpre = (0,) * max(0, inp.ndim-len(padpre)) + padpre
    padpost = (0,) * max(0, inp.ndim-len(padpost)) + padpost
    if inp.dim() != len(padpre) or inp.dim() != len(padpost):
        raise ValueError('Padding length too large')

    # Pad
    mode = ['nocheck'] * max(0, inp.ndim-len(mode)) + mode
    if all(m in ('zero', 'nocheck') for m in mode):
        return _pad_constant(inp, padpre, padpost, value)
    else:
        bound = [getattr(indexing, m) for m in mode]
        return _pad_bound(inp, padpre, padpost, bound)


def _pad_constant(inp, padpre, padpost, value):
    new_shape = [s + pre + post
                 for s, pre, post in zip(inp.shape, padpre, padpost)]
    out = inp.new_full(new_shape, value)
    slicer = [slice(pre, pre + s) for pre, s in zip(padpre, inp.shape)]
    out[tuple(slicer)] = inp
    return out


def _pad_bound(inp, padpre, padpost, bound):
    begin = list(map(lambda x: -x, padpre))
    end = tuple(d+p for d, p in zip(inp.shape, padpost))

    grid = [
        torch.arange(b, e, device=inp.device) for (b, e) in zip(begin, end)
    ]
    mult = [None] * inp.dim()
    for d, n in enumerate(inp.shape):
        grid[d], mult[d] = bound[d](grid[d], n)
    grid = list(meshgrid_list_ij(grid))
    if any(map(torch.is_tensor, mult)):
        for d in range(len(mult)):
            if not torch.is_tensor(mult[d]):
                continue
            for _ in range(d+1, len(mult)):
                mult[d].unsqueeze_(-1)
    mult = prod(mult)
    grid = sub2ind_list(grid, inp.shape)

    out = inp.flatten()[grid]
    if torch.is_tensor(mult) or mult != 1:
        out *= mult
    return out


def ensure_shape(
    inp: Tensor,
    shape: SequenceOrScalar[Optional[int]],
    mode: SequenceOrScalar[BoundLike] = 'constant',
    value: Number = 0,
    side: str = 'post',
    ceil: bool = False
):
    """Pad/crop a tensor so that it has a given shape

    Parameters
    ----------
    inp : tensor
        Input tensor
    shape : SequenceOrScalar[int]
        Output shape
    mode : SequenceOrScalar[BoundLike]
        Boundary mode
    value : scalar, default=0
        Value for mode `'constant'`
    side : "{'pre', 'post', 'both'}"
        Side to crop/pad

    Returns
    -------
    out : tensor
        Padded tensor with shape `shape`

    """
    if isinstance(shape, int):
        shape = [shape]
    shape = list(shape)
    shape = [None] * max(0, inp.ndim - len(shape)) + shape
    if inp.ndim < len(shape):
        inp = inp.reshape((1,) * max(0, len(shape) - inp.ndim) + inp.shape)
    inshape = inp.shape
    shape = [inshape[d] if shape[d] is None else shape[d]
             for d in range(len(shape))]
    ndim = len(shape)

    half = (lambda x: int(math.ceil(x/2))) if ceil else (lambda x: x//2)

    # crop
    if side == 'both':
        crop = [max(0, inshape[d] - shape[d]) for d in range(ndim)]
        index = tuple(slice(half(c), (half(c) - c) or None) for c in crop)
    elif side == 'pre':
        crop = [max(0, inshape[d] - shape[d]) for d in range(ndim)]
        index = tuple(slice(-c or None) for c in crop)
    else:  # side == 'post'
        index = tuple(slice(min(shape[d], inshape[d])) for d in range(ndim))
    inp = inp[index]

    # pad
    pad_size = [max(0, shape[d] - inshape[d]) for d in range(ndim)]
    if side == 'both':
        pad_size = [[half(p), p-half(p)] for p in pad_size]
        pad_size = [q for p in pad_size for q in p]
        side = None
    inp = pad(inp, tuple(pad_size), mode=mode, value=value, side=side)

    return inp


def roll(
    inp: Tensor,
    shifts: SequenceOrScalar[int] = 1,
    dims: Optional[SequenceOrScalar[int]] = None,
    bound: SequenceOrScalar[BoundLike] = 'circular'
):
    r"""Like `torch.roll`, but with any boundary condition

    !!! warning
        When `dims` is `None`, we do not flatten but shift all dimensions.
        This differs from the behavior of `torch.roll` .

    Parameters
    ----------
    inp : tensor
        Input
    shifts : SequenceOrScalar[int]
        Amount by which to roll.
        Positive shifts to the right, negative to the left.
    dims : SequenceOrScalar[int]
        Dimensions to roll.
        By default, shifts apply to all dimensions if a scalar,
        or to the last N if a sequence.
    bound : SequenceOrScalar[BoundLike]
        Boundary condition

    Returns
    -------
    out : tensor
        Rolled tensor

    """
    if dims is None:
        if isinstance(shifts, int):
            dims = list(range(inp.dim()))
        else:
            shifts = ensure_list(shifts)
            dims = list(range(-len(shifts), 0))
    dims = ensure_list(dims)
    shifts = ensure_list(shifts, len(dims))
    bound = map(to_fourier, ensure_list(bound, len(dims)))
    bound = [getattr(indexing, b + '_') for b in bound]

    grid = [torch.arange(n, device=inp.device) for n in inp.shape]
    mult = [1] * inp.dim()
    for d, s, b in zip(dims, shifts, bound):
        grid[d] -= s
        grid[d], mult[d] = b(grid[d], inp.shape[d])
    grid = list(meshgrid_list_ij(grid))
    if any(map(torch.is_tensor, mult)):
        mult = meshgrid_list_ij(mult)
    mult = prod(mult)
    grid = sub2ind_list(grid, inp.shape)

    out = inp.flatten()[grid]
    out *= mult
    return out



"""
This module defines names and aliases for different boundary conditions,
as well as tools to convert between different naming conventions.

Classes
-------
BoundType
    Enum type for bounds

Attributes
----------
BoundLike
    A type hint for any boundary type
SequenceOrScalar
    A type hint for values or sequences of values

Functions
---------
to_enum
    Convert any boundary type to `BoundType`
to_int
    Convert any boundary type to `BoundType`-based integer values
to_fourier
    Convert any boundary type to discrete transforms
to_scipy
    Convert any boundary type to `scipy.ndimage` convention
to_torch
    Convert any boundary type to `torch.grid_sample` convention

"""  # noqa: E501
__all__ = [
    'BoundType',
    'BoundLike',
    'SequenceOrScalar',
    'to_enum',
    'to_int',
    'to_fourier',
    'to_scipy',
    'to_torch',
]
from enum import Enum
from typing import Union, Sequence, TypeVar


class BoundType(Enum):
    """
    An Enum type that maps boundry modes of any convention to a
    unique set of values.

    ```python
    class BoundType(Enum):
        zero = zeros = constant = gridconstant = 0
        replicate = repeat = nearest = border = edge = 1
        dct1 = mirror = 2
        dct2 = reflect = reflection = gridmirror = neumann = 3
        dst1 = antimirror = 4
        dst2 = antireflect = dirichlet = 5
        dft = fft = wrap = gridwrap = circular = circulant = 6
        nocheck = -1
    ```
    """
    zero = zeros = constant = gridconstant = 0
    replicate = repeat = nearest = border = edge = 1
    dct1 = mirror = 2
    dct2 = reflect = reflection = gridmirror = neumann = 3
    dst1 = antimirror = 4
    dst2 = antireflect = dirichlet = 5
    dft = fft = wrap = gridwrap = circular = circulant = 6
    nocheck = -1


T = TypeVar('T')

SequenceOrScalar = Union[T, Sequence[T]]
"""Either an element or type `T`, or a sequence of elements of type `T`."""

BoundLike = Union[BoundType, str, int]
"""
A boundary mode.

Most conventions are handled (numpy, scipy, torch, see below). Can be one of:

0. `zero`, `zeros`, `constant` or `gridconstant`;
1. `replicate`, `nearest`, `border` or `edge`;
2. `dct1` or `mirror`;
3. `dct2`, `reflect`, `reflection`, `gridmirror` or `neumann`;
4. `dst1` or `antimirror`;
5. `dst2`, `antireflect` or `dirichlet`;
6. `dft`, `fft`, `wrap`, `gridwrap`, `circular` or `circulant`.

Each of these modes can be a [`BoundType`][bounds.types.BoundType] value
(e.g., `BoundType.mirror`), or its string representation (e.g., `"mirror"`).

The aliases `dft`, `dct1`, `dct2`, `dst1` and `dst2` exist because these
boundary modes correspond to the implicit boundary conditions of each of
these frequency transform:

- [Discrete Fourier Transform (DFT)](https://w.wiki/92by)
- [Discrete Cosine Transform I & II (DCT-I, DCT-II)](https://w.wiki/AQEt)
- [Discrete Sine Transform I & II (DST-I, DST-II)](https://w.wiki/ATnn)

The reason why so many aliases are supported is that there is no common
convention across python packages to name boundary conditions.
This table contains an extensive list of aliases:

+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| Fourier | SciPy `ndimage` | Numpy `pad`    | PyTorch `pad` | PyTorch `grid_sample` | Other        | Description                                    |
+=========+=================+================+===============+=======================+==============+================================================+
|         | nearest         | edge           | border        | replicate             | repeat       | <code> a  a &#124; a b c d &#124;  d  d</code> |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
|         | constant,<br /> | constant       | constant      | zeros                 | zero         | <code> 0  0 &#124; a b c d &#124;  0  0</code> |
|         | grid-constant   |                |               |                       |              |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| dct1    | mirror          | reflect        | reflect       | reflection<br />      |              | <code> c  b &#124; a b c d &#124;  c  b</code> |
|         |                 |                |               | (`False`)             |              |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| dct2    | reflect,<br />  | symmetric      |               | reflection<br />      | neumann      | <code> b  a &#124; a b c d &#124;  d  c</code> |
|         | grid-mirror     |                |               | (`True`)              |              |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| dst1    |                 |                |               |                       | antimirror   | <code>-a  0 &#124; a b c d &#124;  0 -d</code> |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| dst2    |                 |                |               |                       | antireflect, | <code>-b -a &#124; a b c d &#124; -d -c</code> |
|         |                 |                |               |                       | dirichlet    |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
| dft     | grid-wrap       | wrap           | circular      |                       | circulant    | <code> c  d &#124; a b c d &#124;  a  b</code> |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
|         | wrap            |                |               |                       |              | <code> c  d &#124; a b c d &#124;  b  c</code> |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
|         |                 | linear_ramp    |               |                       |              |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+
|         |                 | minimum,<br /> |               |                       |              |                                                |
|         |                 | maximum,<br /> |               |                       |              |                                                |
|         |                 | mean,<br />    |               |                       |              |                                                |
|         |                 | median         |               |                       |              |                                                |
+---------+-----------------+----------------+---------------+-----------------------+--------------+------------------------------------------------+

Some of these conventions are inconsistant with each other. For example
`"wrap"` in `scipy.ndimage` is different from `"wrap"` in `numpy.pad`,
which corresponds to `"grid-wrap"` in `scipy.ndimage`. Also, `"reflect"`
in `numpy.pad` and `torch.pad` is different from `"reflect"` in `scipy.ndimage`,
which correspond to `"symmetric"` in `numpy.pad`.
"""  # noqa: E501


bounds_fourier = ('replicate', 'zero', 'dct2', 'dct1', 'dst2', 'dst1', 'dft')
bounds_scipy = ('nearest', 'constant', 'reflect', 'mirror', 'wrap')
bounds_torch = ('nearest', 'zeros', 'reflection')
bounds_torch_pad = ('border', 'constant', 'reflect', 'circular')
bounds_other = ('repeat', 'neumann', 'circular', 'circulant',
                'antireflect', 'dirichlet', 'antimirror', 'fft')
enum_bounds = (BoundType.zero, BoundType.repeat, BoundType.dct1,
               BoundType.dct2, BoundType.dst1, BoundType.dst2, BoundType.dft)
int_bounds = tuple(range(7))


zero_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.zero]
rept_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.repeat]
dct1_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.dct1]
dct2_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.dct2]
dst1_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.dst1]
dst2_bounds = [b for b in BoundType.__members__.keys()
               if getattr(BoundType, b) == BoundType.dst2]
dft_bounds = [b for b in BoundType.__members__.keys()
              if getattr(BoundType, b) == BoundType.dft]


def to_enum(bound: SequenceOrScalar[BoundLike]) -> SequenceOrScalar[BoundType]:
    """Convert boundary type to enum type.

    !!! note "See also"
        * [`to_fourier`][bounds.types.to_fourier]
        * [`to_scipy`][bounds.types.to_scipy]
        * [`to_torch`][bounds.types.to_torch]
        * [`to_int`][bounds.types.to_int]

    Parameters
    ----------
    bound : SequenceOrScalar[BoundLike]
        Boundary condition in any convention

    Returns
    -------
    bound : SequenceOrScalar[BoundType]
        Boundary condition

    """
    intype = type(bound)
    if not isinstance(bound, (list, tuple)):
        bound = [bound]
    obound = []
    for b in bound:
        if isinstance(b, str):
            b = b.lower()
        if b in (*zero_bounds, BoundType.zero, 0):
            obound.append(BoundType.zero)
        elif b in (*rept_bounds, BoundType.border, 1):
            obound.append(BoundType.replicate)
        elif b in (*dct1_bounds, BoundType.dct1, 2):
            obound.append(BoundType.dct1)
        elif b in (*dct2_bounds, BoundType.dct2, 3):
            obound.append(BoundType.dct2)
        elif b in (*dst1_bounds, BoundType.dst1, 4):
            obound.append(BoundType.dst1)
        elif b in (*dst2_bounds, BoundType.dst2, 5):
            obound.append(BoundType.dst2)
        elif b in (*dft_bounds, BoundType.dft, 6):
            obound.append(BoundType.dft)
        elif b in ('nocheck', BoundType.nocheck, -1):
            obound.append(BoundType.nocheck)
        else:
            raise ValueError(f'Unknown boundary condition {b}')
    if issubclass(intype, (list, tuple)):
        obound = intype(obound)
    else:
        obound = obound[0]
    return obound


def to_int(bound: SequenceOrScalar[BoundLike]) -> SequenceOrScalar[int]:
    """Convert boundary type to enum integer.

    !!! note "See also"
        * [`to_enum`][bounds.types.to_enum]
        * [`to_fourier`][bounds.types.to_fourier]
        * [`to_scipy`][bounds.types.to_scipy]
        * [`to_torch`][bounds.types.to_torch]

    Parameters
    ----------
    bound : SequenceOrScalar[BoundLike]
        Boundary condition in any convention

    Returns
    -------
    bound : SequenceOrScalar[{0..6}]
        Boundary condition

    """
    bound = to_enum(bound)
    if isinstance(bound, (list, tuple)):
        bound = type(bound)(map(lambda x: x.value, bound))
    else:
        bound = bound.value
    return bound


def to_fourier(bound: SequenceOrScalar[BoundLike]) -> SequenceOrScalar[str]:
    """Convert boundary type to discrete transforms.

    !!! note "See also"
        * [`to_enum`][bounds.types.to_enum]
        * [`to_scipy`][bounds.types.to_scipy]
        * [`to_torch`][bounds.types.to_torch]
        * [`to_int`][bounds.types.to_int]

    Parameters
    ----------
    bound : SequenceOrScalar[BoundLike]
        Boundary condition in any convention

    Returns
    -------
    bound : SequenceOrScalar[{'replicate', 'zero', 'dct2', 'dct1', 'dst2', 'dst1', 'dft'}]
        Boundary condition in terms of discrete transforms

    """  # noqa: E501
    intype = type(bound)
    if not isinstance(bound, (list, tuple)):
        bound = [bound]
    obound = []
    for b in bound:
        if isinstance(b, str):
            b = b.lower()
        if b in (*zero_bounds, BoundType.zero, 0):
            obound.append('zero')
        elif b in (*rept_bounds, BoundType.border, 1):
            obound.append('replicate')
        elif b in (*dct1_bounds, BoundType.dct1, 2):
            obound.append('dct1')
        elif b in (*dct2_bounds, BoundType.dct2, 3):
            obound.append('dct2')
        elif b in (*dst1_bounds, BoundType.dst1, 4):
            obound.append('dst1')
        elif b in (*dst2_bounds, BoundType.dst2, 5):
            obound.append('dst2')
        elif b in (*dft_bounds, BoundType.dft, 6):
            obound.append('dft')
        elif b in ('nocheck', BoundType.nocheck, -1):
            obound.append('nocheck')
        else:
            raise ValueError(f'Unknown boundary condition {b}')
    if issubclass(intype, (list, tuple)):
        obound = intype(obound)
    else:
        obound = obound[0]
    return obound


def to_scipy(bound: SequenceOrScalar[BoundLike]) -> SequenceOrScalar[str]:
    """Convert boundary type to SciPy's convention.

    !!! note "See also"
        * [`to_enum`][bounds.types.to_enum]
        * [`to_fourier`][bounds.types.to_fourier]
        * [`to_torch`][bounds.types.to_torch]
        * [`to_int`][bounds.types.to_int]

    Parameters
    ----------
    bound : SequenceOrScalar[BoundLike]
        Boundary condition in any convention

    Returns
    -------
    bound : SequenceOrScalar[{'border', 'constant', 'reflect', 'mirror', 'wrap'}]
        Boundary condition in SciPy's convention

    """  # noqa: E501
    intype = type(bound)
    if not isinstance(bound, (list, tuple)):
        bound = [bound]
    obound = []
    for b in bound:
        if isinstance(b, str):
            b = b.lower()
        if b in (*zero_bounds, BoundType.zero, 0):
            obound.append('constant')
        elif b in (*rept_bounds, BoundType.border, 1):
            obound.append('border')
        elif b in (*dct1_bounds, BoundType.dct1, 2):
            obound.append('mirror')
        elif b in (*dct2_bounds, BoundType.dct2, 3):
            obound.append('reflect')
        elif b in (*dst1_bounds, BoundType.dst1, 4):
            raise ValueError(f'Boundary condition {b} not available in SciPy.')
        elif b in (*dst2_bounds, BoundType.dst2, 5):
            raise ValueError(f'Boundary condition {b} not available in SciPy.')
        elif b in (*dft_bounds, BoundType.dft, 6):
            obound.append('wrap')
        elif b in ('nocheck', BoundType.nocheck, -1):
            raise ValueError(f'Boundary condition {b} not available in SciPy.')
        else:
            raise ValueError(f'Unknown boundary condition {b}')
    if issubclass(intype, (list, tuple)):
        obound = intype(obound)
    else:
        obound = obound[0]
    return obound


def to_torch(bound: SequenceOrScalar[BoundLike]) -> SequenceOrScalar[str]:
    """Convert boundary type to PyTorch's convention.

    !!! note "See also"
        * [`to_enum`][bounds.types.to_enum]
        * [`to_fourier`][bounds.types.to_fourier]
        * [`to_scipy`][bounds.types.to_scipy]
        * [`to_int`][bounds.types.to_int]

    Parameters
    ----------
    bound : SequenceOrScalar[BoundLike]
        Boundary condition in any convention

    Returns
    -------
    bound : SequenceOrScalar[({'nearest', 'zero', 'reflection'}, bool)]
        The first element is the boundary condition in PyTorchs's
        convention, and the second element is the value of `align_corners`.

    """
    intype = type(bound)
    if not isinstance(bound, (list, tuple)):
        bound = [bound]
    obound = []
    for b in bound:
        if isinstance(b, str):
            b = b.lower()
        if b in (*zero_bounds, BoundType.zero, 0):
            obound.append(('zero', None))
        elif b in (*rept_bounds, BoundType.border, 1):
            obound.append(('nearest', None))
        elif b in (*dct1_bounds, BoundType.dct1, 2):
            obound.append(('reflection', False))
        elif b in (*dct2_bounds, BoundType.dct2, 3):
            obound.append(('reflection', True))
        elif b in (*dst1_bounds, BoundType.dst1, 4):
            raise ValueError(f'Boundary condition {b} not available in Torch.')
        elif b in (*dst2_bounds, BoundType.dst2, 5):
            raise ValueError(f'Boundary condition {b} not available in Torch.')
        elif b in (*dft_bounds, BoundType.dft, 6):
            raise ValueError(f'Boundary condition {b} not available in Torch.')
        elif b in ('nocheck', BoundType.nocheck, -1):
            raise ValueError(f'Boundary condition {b} not available in Torch.')
        else:
            raise ValueError(f'Unknown boundary condition {b}')
    if issubclass(intype, (list, tuple)):
        obound = intype(obound)
    else:
        obound = obound[0]
    return obound



"""

Modules
-------
indexing
    Functions that wrap out-of-bound indices back in-bounds,
    according to some boundary condition.
padding
    Reimplements [`torch.nn.functional.pad`][] and [`torch.roll`][]
    with a larger set of boundary conditions.
realtransforms
    Implements the discrete sine and cosine transforms, variants I, II, III.
types
    Defines names and aliases for different boundary conditions,
    as well as tools to convert between different naming conventions.

Functions
---------
pad
    Pad a tensor
roll
    Roll a tensor
ensure_shape
    Pad/crop a tensor so that it has a given shape
to_enum
    Convert any boundary type to `BoundType`
to_int
    Convert any boundary type to `BoundType`-based integer values
to_fourier
    Convert any boundary type to discrete transforms
to_scipy
    Convert any boundary type to `scipy.ndimage` convention
to_torch
    Convert any boundary type to `torch.grid_sample` convention
dct
    One-dimensional Discrete Cosine Transform (DCT)
dst
    One-dimensional Discrete Sine Transform (DST)
idct
    One-dimensional Inverse Discrete Cosine Transform (DCT)
idst
    One-dimensional Inverse Discrete Sine Transform (DST)
dctn
    N-dimensional Discrete Cosine Transform (IDCT)
dstn
    N-dimensional Discrete Sine Transform (IDST)
idctn
    N-dimensional Inverse Discrete Cosine Transform (IDCT)
idstn
    N-dimensional Inverse Discrete Sine Transform (IDST)

Classes
-------
BoundType
    Enum type for bounds

Attributes
----------
BoundLike
    A type hint for any boundary type
SequenceOrScalar
    A type hint for values or sequences of values
"""
import torch as _torch

from .padding import *      # noqa: F401, F403
from .types import *        # noqa: F401, F403

from . import indexing      # noqa: F401
from . import padding       # noqa: F401
from . import types         # noqa: F401

__all__ = [
    'indexing',
    'padding',
    'types',
]
__all__ += padding.__all__
__all__ += types.__all__


if hasattr(getattr(_torch, 'fft', None), 'fft'):

    from .realtransforms import *        # noqa: F401, F403
    from . import realtransforms         # noqa: F401

    __all__ += ['realtransforms']
    __all__ += realtransforms.__all__


from . import _version  # noqa: E402
__version__ = _version.get_versions()['version']



import os
import torch
from torch import Tensor
from types import GeneratorType as generator
from typing import List, Any, Optional, Union


def ensure_list(x: Any, length: Optional[int] = None, crop: bool = True,
                **kwargs) -> List:
    """
    Ensure that an object is a list

    The output list is of length at least `length`.
    When `crop` is `True`, its length is also at most `length`.
    If needed, the last value is replicated, unless `default` is provided.

    If x is a list, nothing is done (no copy triggered).
    If it is a tuple, range, or generator, it is converted to a list.
    Otherwise, it is placed inside a list.
    """
    if not isinstance(x, (list, tuple, range, generator)):
        x = [x]
    elif not isinstance(x, list):
        x = list(x)
    if length and len(x) < length:
        default = [kwargs.get('default', x[-1] if x else None)]
        x += default * (length - len(x))
    if length and crop:
        x = x[:length]
    return x


def make_vector(
    input: Tensor,
    n: Optional[int] = None,
    crop: bool = True,
    *args,
    dtype: Optional[torch.dtype] = None,
    device: Optional[Union[str, torch.device]] = None,
    **kwargs
):
    """Ensure that the input is a (tensor) vector and pad/crop if necessary.

    Parameters
    ----------
    input : scalar or sequence or generator
        Input argument(s).
    n : int, optional
        Target length.
    crop : bool, default=True
        Crop input sequence if longer than `n`.
    default : optional
        Default value to pad with.
        If not provided, replicate the last value.
    dtype : torch.dtype, optional
        Output data type.
    device : torch.device, optional
        Output device

    Returns
    -------
    output : tensor
        Output vector.

    """
    input = torch.as_tensor(input, dtype=dtype, device=device).flatten()
    if n is None:
        return input
    if n is not None and input.numel() >= n:
        return input[:n] if crop else input
    if args:
        default = args[0]
    elif 'default' in kwargs:
        default = kwargs['default']
    else:
        default = input[-1]
    output = input.new_full([n], default)
    output[:len(input)] = input
    return output


def prod(sequence, inplace=False):
    """Perform the product of a sequence of elements.

    Parameters
    ----------
    sequence : any object that implements `__iter__`
        Sequence of elements for which the `__mul__` operator is defined.
    inplace : bool, default=False
        Perform the product inplace (using `__imul__` instead of `__mul__`).

    Returns
    -------
    product :
        Product of the elements in the sequence.

    """
    accumulate = None
    for elem in sequence:
        if accumulate is None:
            accumulate = elem
        elif inplace:
            accumulate *= elem
        else:
            accumulate = accumulate * elem
    return accumulate


def _compare_versions(version1, mode, version2):
    for v1, v2 in zip(version1, version2):
        if mode in ('gt', '>'):
            if v1 > v2:
                return True
            elif v1 < v2:
                return False
        elif mode in ('ge', '>='):
            if v1 > v2:
                return True
            elif v1 < v2:
                return False
        elif mode in ('lt', '<'):
            if v1 < v2:
                return True
            elif v1 > v2:
                return False
        elif mode in ('le', '<='):
            if v1 < v2:
                return True
            elif v1 > v2:
                return False
    if mode in ('gt', 'lt', '>', '<'):
        return False
    else:
        return True


def torch_version(mode, version):
    """Check torch version

    Parameters
    ----------
    mode : {'<', '<=', '>', '>='}
    version : tuple[int]

    Returns
    -------
    True if "torch.version <mode> version"

    """
    current_version, *cuda_variant = torch.__version__.split('+')
    major, minor, patch, *_ = current_version.split('.')
    # strip alpha tags
    for x in 'abcdefghijklmnopqrstuvwxy':
        if x in patch:
            patch = patch[:patch.index(x)]
    current_version = (int(major), int(minor), int(patch))
    version = ensure_list(version)
    return _compare_versions(current_version, mode, version)


# In torch < 1.6, div applied to integer tensor performed a floor_divide
# In torch > 1.6, it performs a true divide.
# Floor division must be done using `floor_divide`, but it was buggy
# until torch 1.13 (it was doing a trunc divide instead of a floor divide).
# There was at some point a deprecation warning for floor_divide, but it
# seems to have been lifted afterwards. In torch >= 1.13, floor_divide
# performs a correct floor division.
# Since we only apply floor_divide ot positive values, we are fine.
if torch_version('>=', [1, 8]):
    @torch.jit.script
    def floor_div(x, y) -> torch.Tensor:
        return torch.div(x, y, rounding_mode='floor')

    @torch.jit.script
    def floor_div_int(x, y: int) -> torch.Tensor:
        return torch.div(x, y, rounding_mode='floor')
elif torch_version('<', (1, 6)):
    floor_div = floor_div_int = torch.div
else:
    @torch.jit.script
    def floor_div(x, y) -> torch.Tensor:
        return torch.div(x, y, rounding_mode='floor')

    @torch.jit.script
    def floor_div_int(x, y: int) -> torch.Tensor:
        return torch.div(x, y, rounding_mode='floor')

    floor_div = floor_div_int = torch.floor_divide

#     @torch.jit.script
#     def floor_div(x, y) -> torch.Tensor:
#         return (x / y).floor_()
#     @torch.jit.script
#     def floor_div_int(x, y: int) -> torch.Tensor:
#         return (x / y).floor_()


if torch_version('>=', (1, 10)):
    # torch >= 1.10
    # -> use `indexing` keyword

    if not int(os.environ.get('PYTORCH_JIT', '1')):
        # JIT deactivated -> torch.meshgrid takes an unpacked list of tensors

        @torch.jit.script
        def meshgrid_list_ij(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(*tensors, indexing='ij'))

        @torch.jit.script
        def meshgrid_list_xy(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(*tensors, indexing='xy'))

    else:
        # JIT activated -> torch.meshgrid takes a packed list of tensors

        @torch.jit.script
        def meshgrid_list_ij(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(tensors, indexing='ij'))

        @torch.jit.script
        def meshgrid_list_xy(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(tensors, indexing='xy'))

else:
    # torch < 1.10
    # -> implement "xy" mode manually

    if not int(os.environ.get('PYTORCH_JIT', '1')):
        # JIT deactivated -> torch.meshgrid takes an unpacked list of tensors

        @torch.jit.script
        def meshgrid_list_ij(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(tensors))

        @torch.jit.script
        def meshgrid_list_xy(tensors: List[Tensor]) -> List[Tensor]:
            grid = list(torch.meshgrid(*tensors))
            if len(grid) > 1:
                grid[0] = grid[0].transpose(0, 1)
                grid[1] = grid[1].transpose(0, 1)
            return grid

    else:
        # JIT activated -> torch.meshgrid takes a packed list of tensors

        @torch.jit.script
        def meshgrid_list_ij(tensors: List[Tensor]) -> List[Tensor]:
            return list(torch.meshgrid(tensors))

        @torch.jit.script
        def meshgrid_list_xy(tensors: List[Tensor]) -> List[Tensor]:
            grid = list(torch.meshgrid(tensors))
            if len(grid) > 1:
                grid[0] = grid[0].transpose(0, 1)
                grid[1] = grid[1].transpose(0, 1)
            return grid


@torch.jit.script
def reverse_list_int(x: List[int]) -> List[int]:
    """TorchScript equivalent to `x[::-1]`"""
    if len(x) == 0:
        return x
    return [x[i] for i in range(-1, -len(x)-1, -1)]


@torch.jit.script
def cumprod_list_int(x: List[int], reverse: bool = False,
                     exclusive: bool = False) -> List[int]:
    """Cumulative product of elements in the list

    Parameters
    ----------
    x : list[int]
        List of integers
    reverse : bool
        Cumulative product from right to left.
        Else, cumulative product from left to right (default).
    exclusive : bool
        Start series from 1.
        Else start series from first element (default).

    Returns
    -------
    y : list[int]
        Cumulative product

    """
    if len(x) == 0:
        lx: List[int] = []
        return lx
    if reverse:
        x = reverse_list_int(x)

    x0 = 1 if exclusive else x[0]
    lx = [x0]
    all_x = x[:-1] if exclusive else x[1:]
    for x1 in all_x:
        x0 = x0 * x1
        lx.append(x0)
    if reverse:
        lx = reverse_list_int(lx)
    return lx


@torch.jit.script
def sub2ind_list(subs: List[Tensor], shape: List[int]):
    """Convert sub indices (i, j, k) into linear indices.

    The rightmost dimension is the most rapidly changing one
    -> if shape == [D, H, W], the strides are therefore [H*W, W, 1]

    Parameters
    ----------
    subs : (D,) list[tensor]
        List of sub-indices. The first dimension is the number of dimension.
        Each element should have the same number of elements and shape.
    shape : (D,) list[int]
        Size of each dimension. Its length should be the same as the
        first dimension of ``subs``.

    Returns
    -------
    ind : (...) tensor
        Linear indices
    """
    ind = subs[-1]
    subs = subs[:-1]
    ind = ind.clone()
    stride = cumprod_list_int(shape[1:], reverse=True, exclusive=False)
    for i, s in zip(subs, stride):
        ind += i * s
    return ind



# copied and adapted from
#   https://github.com/cupy/cupy/blob/v12.0.0/cupyx/scipy/fft/_realtransforms.py
#   https://github.com/cupy/cupy/blob/v12.0.0/LICENSE
#
# NOTE (YB 2023/05/12)
#   Scipy's "orthogonal" DST is not orthogonal
#   I've modified it to make it properly orthogonal, but it means my
#   version is not exactly equivalent to scipy/cupy anymore.
#   Scipy's buggy behaviour can be recovered with the option
#   `norm="ortho_scipy"`.
#
# NOTE (YB 2024/02/23)
#   Reported in https://github.com/scipy/scipy/issues/18452
#   Fixed by https://github.com/scipy/scipy/pull/18613      (scipy)
#   Fixed by https://github.com/cupy/cupy/pull/8058         (cupy)
"""Real-to-real transforms

cuFFT does not implement real-to-real FFTs. This module implements forward
and inverse DCT-II and DCT-III transforms using FFTs.

A length N DCT can be computed using a length N FFT and some additional
multiplications and reordering of entries.

The approach taken here is based on the work in [1]_, [2]_ and is discussed in
the freely-available online resources [3]_, [4]_.

The implementation here follows that approach with only minor modification to
match the normalization conventions in SciPy.

The modifications to turn a type II or III DCT to a DST were implemented as
described in [5]_.

.. [1] J. Makhoul, "A fast cosine transform in one and two dimensions," in
    IEEE Transactions on Acoustics, Speech, and Signal Processing, vol. 28,
    no. 1, pp. 27-34, February 1980.

.. [2] M.J. Narasimha and A.M. Peterson, “On the computation of the discrete
    cosine  transform,” IEEE Trans. Commun., vol. 26, no. 6, pp. 934–936, 1978.

.. [3] http://fourier.eng.hmc.edu/e161/lectures/dct/node2.html

.. [4] https://dsp.stackexchange.com/questions/2807/fast-cosine-transform-via-fft  # noqa

.. [5] X. Shao, S. G. Johnson. Type-II/III DCT/DST algorithms with reduced
    number of arithmetic operations, Signal Processing, Volume 88, Issue 6,
    pp. 1553-1564, 2008.
"""
import math
import numbers
import operator
import torch
from torch import fft as _fft


__all__ = ['dct', 'dctn', 'dst', 'dstn', 'idct', 'idctn', 'idst', 'idstn']


def dct(x, type=2, n=None, dim=-1, norm=None, overwrite_x=False):
    """Return the Discrete Cosine Transform of an array, x.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DCT (see Notes). Default type is 2. Currently CuPy only
        supports types 2 and 3.
    n : int, optional:
        Length of the transform.  If ``n < x.shape[dim]``, `x` is
        truncated. If ``n > x.shape[dim]``, `x` is zero-padded.
        The default results in ``n = x.shape[dim]``.
    dim : int, optional
        Axis along which the dct is computed; the default is over the
        last dim (i.e., ``dim=-1``).
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    y : torch.Tensor
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.dct`

    Notes
    -----
    For a single dimension array ``x``, ``dct(x, norm='ortho')`` is equal
    to MATLAB ``dct(x)``.

    For ``norm="ortho"`` both the `dct` and `idct` are scaled by the same
    overall factor in both directions. By default, the transform is also
    orthogonalized which for types 1, 2 and 3 means the transform definition is
    modified to give orthogonality of the DCT matrix (see below).

    For ``norm="backward"``, there is no scaling on `dct` and the `idct` is
    scaled by ``1/N`` where ``N`` is the "logical" size of the DCT. For
    ``norm="forward"`` the ``1/N`` normalization is applied to the forward
    `dct` instead and the `idct` is unnormalized.

    CuPy currently only supports DCT types 2 and 3. 'The' DCT generally
    refers to DCT type 2, and 'the' Inverse DCT generally refers to DCT
    type 3 [1]_. See the :func:`scipy.fft.dct` documentation for a full
    description of each type.

    References
    ----------
    .. [1] Wikipedia, "Discrete cosine transform",
           https://en.wikipedia.org/wiki/Discrete_cosine_transform

    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = dct(x.real, type, n, dim, norm, overwrite_x)
        out = out + 1j * dct(x.imag, type, n, dim, norm, overwrite_x)
        return out

    x = _promote_dtype(x)

    kwargs = dict(n=n, dim=dim, norm=norm, forward=True)
    if type == 2:
        return _dct_or_dst_type2(x, dst=False, **kwargs)
    elif type == 3:
        return _dct_or_dst_type3(x, dst=False, **kwargs)
    elif type == 1:
        return _dct_type1(x, **kwargs)
    elif type == 4:
        raise NotImplementedError(
            'Only DCT I, II and III have been implemented.'
        )
    else:
        raise ValueError('invalid DCT type')


def dst(x, type=2, n=None, dim=-1, norm=None, overwrite_x=False):
    """Return the Discrete Sine Transform of an array, x.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DST (see Notes). Default type is 2.
    n : int, optional
        Length of the transform.  If ``n < x.shape[dim]``, `x` is
        truncated.  If ``n > x.shape[dim]``, `x` is zero-padded. The
        default results in ``n = x.shape[dim]``.
    dim : int, optional
        Axis along which the dst is computed; the default is over the
        last dim (i.e., ``dim=-1``).
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    dst : torch.Tensor
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.dst`

    Notes
    -----

    For ``norm="ortho"`` both the `dst` and `idst` are scaled by the same
    overall factor in both directions. By default, the transform is also
    orthogonalized which for types 2 and 3 means the transform definition is
    modified to give orthogonality of the DST matrix (see below).

    For ``norm="backward"``, there is no scaling on the `dst` and the `idst` is
    scaled by ``1/N`` where ``N`` is the "logical" size of the DST.

    See the :func:`scipy.fft.dst` documentation for a full description of each
    type. CuPy currently only supports DST types 2 and 3.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = dst(x.real, type, n, dim, norm, overwrite_x)
        out = out + 1j * dst(x.imag, type, n, dim, norm, overwrite_x)
        return out

    x = _promote_dtype(x)

    kwargs = dict(n=n, dim=dim, norm=norm, forward=True)
    if type == 2:
        return _dct_or_dst_type2(x, dst=True, **kwargs)
    elif type == 3:
        return _dct_or_dst_type3(x, dst=True, **kwargs)
    elif type == 1:
        return _dst_type1(x, **kwargs)
    elif type == 4:
        raise NotImplementedError(
            'Only DST I, II and III have been implemented.'
        )
    else:
        raise ValueError('invalid DST type')


def idct(x, type=2, n=None, dim=-1, norm=None, overwrite_x=False):
    """Return the Inverse Discrete Cosine Transform of an array, x.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DCT (see Notes). Default type is 2.
    n : int, optional
        Length of the transform.  If ``n < x.shape[dim]``, `x` is
        truncated.  If ``n > x.shape[dim]``, `x` is zero-padded. The
        default results in ``n = x.shape[dim]``.
    dim : int, optional
        Axis along which the idct is computed; the default is over the
        last dim (i.e., ``dim=-1``).
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    idct : torch.Tensor
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.idct`

    Notes
    -----
    For a single dimension array `x`, ``idct(x, norm='ortho')`` is equal to
    MATLAB ``idct(x)``.

    For ``norm="ortho"`` both the `dct` and `idct` are scaled by the same
    overall factor in both directions. By default, the transform is also
    orthogonalized which for types 1, 2 and 3 means the transform definition is
    modified to give orthogonality of the IDCT matrix (see `dct` for the full
    definitions).

    'The' IDCT is the IDCT-II, which is the same as the normalized DCT-III
    [1]_. See the :func:`scipy.fft.dct` documentation for a full description of
    each type. CuPy currently only supports DCT types 2 and 3.

    References
    ----------
    .. [1] Wikipedia, "Discrete sine transform",
           https://en.wikipedia.org/wiki/Discrete_sine_transform
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = idct(x.real, type, n, dim, norm, overwrite_x)
        out = out + 1j * idct(x.imag, type, n, dim, norm, overwrite_x)
        return out

    x = _promote_dtype(x)

    kwargs = dict(n=n, dim=dim, norm=norm, forward=False)
    if type == 2:
        # DCT-III is the inverse of DCT-II
        return _dct_or_dst_type3(x, dst=False, **kwargs)
    elif type == 3:
        # DCT-II is the inverse of DCT-III
        return _dct_or_dst_type2(x, dst=False, **kwargs)
    elif type == 1:
        # DCT-I is the inverse of DCT-I
        return _dct_type1(x, **kwargs)
    elif type == 4:
        raise NotImplementedError(
            'Only DCT I, II and III have been implemented.'
        )
    else:
        raise ValueError('invalid DCT type')


def idst(x, type=2, n=None, dim=-1, norm=None, overwrite_x=False):
    """Return the Inverse Discrete Sine Transform of an array, x.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DST (see Notes). Default type is 2.
    n : int, optional
        Length of the transform. If ``n < x.shape[dim]``, `x` is
        truncated.  If ``n > x.shape[dim]``, `x` is zero-padded. The
        default results in ``n = x.shape[dim]``.
    dim : int, optional
        Axis along which the idst is computed; the default is over the
        last dim (i.e., ``dim=-1``).
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    idst : torch.Tensor of real
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.idst`

    Notes
    -----
    For full details of the DST types and normalization modes, as well as
    references, see :func:`scipy.fft.dst`.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = idst(x.real, type, n, dim, norm, overwrite_x)
        out = out + 1j * idst(x.imag, type, n, dim, norm, overwrite_x)
        return out

    x = _promote_dtype(x)

    kwargs = dict(n=n, dim=dim, norm=norm, forward=False)
    if type == 2:
        # DST-III is the inverse of DST-II
        return _dct_or_dst_type3(x, dst=True, **kwargs)
    elif type == 3:
        # DST-II is the inverse of DST-III
        return _dct_or_dst_type2(x, dst=True, **kwargs)
    elif type == 1:
        # DST-I is the inverse of DST-I
        return _dst_type1(x, **kwargs)
    elif type == 4:
        raise NotImplementedError(
            'Only DST I, II and III have been implemented.'
        )
    else:
        raise ValueError('invalid DST type')


def dctn(x, type=2, s=None, dim=None, norm=None, overwrite_x=False):
    """Compute a multidimensional Discrete Cosine Transform.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DCT (see Notes). Default type is 2.
    s : int or array_like of ints or None, optional
        The shape of the result. If both `s` and `dim` (see below) are None,
        `s` is ``x.shape``; if `s` is None but `dim` is not None, then `s` is
        ``numpy.take(x.shape, dim, dim=0)``.
        If ``s[i] > x.shape[i]``, the ith dimension is padded with zeros.
        If ``s[i] < x.shape[i]``, the ith dimension is truncated to length
        ``s[i]``.
        If any element of `s` is -1, the size of the corresponding dimension of
        `x` is used.
    dim : int or array_like of ints or None, optional
        Axes over which the DCT is computed. If not given, the last ``len(s)``
        dimensions are used, or all dimensions if `s` is also not specified.
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    y : torch.Tensor of real
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.dctn`

    Notes
    -----
    For full details of the DCT types and normalization modes, as well as
    references, see `dct`.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = dctn(x.real, type, s, dim, norm, overwrite_x)
        out = torch.complex(out, dctn(x.imag, type, s, dim, norm, overwrite_x))
        return out

    shape, dim = _init_nd_shape_and_dims(x, s, dim)
    x = _promote_dtype(x)

    if len(dim) == 0:
        return x

    kwargs = dict(type=type, norm=norm, overwrite_x=overwrite_x)
    for n, dim1 in zip(shape, dim):
        x = dct(x, n=n, dim=dim1, **kwargs)
    return x


def idctn(x, type=2, s=None, dim=None, norm=None, overwrite_x=False):
    """Compute a multidimensional Discrete Cosine Transform.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DCT (see Notes). Default type is 2.
    s : int or array_like of ints or None, optional
        The shape of the result. If both `s` and `dim` (see below) are None,
        `s` is ``x.shape``; if `s` is None but `dim` is not None, then `s` is
        ``numpy.take(x.shape, dim, dim=0)``.
        If ``s[i] > x.shape[i]``, the ith dimension is padded with zeros.
        If ``s[i] < x.shape[i]``, the ith dimension is truncated to length
        ``s[i]``.
        If any element of `s` is -1, the size of the corresponding dimension of
        `x` is used.
    dim : int or array_like of ints or None, optional
        Axes over which the IDCT is computed. If not given, the last ``len(s)``
        dimensions are used, or all dimensions if `s` is also not specified.
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    y : torch.Tensor of real
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.idctn`

    Notes
    -----
    For full details of the IDCT types and normalization modes, as well as
    references, see :func:`scipy.fft.idct`.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = idctn(x.real, type, s, dim, norm, overwrite_x)
        out = out + 1j * idctn(x.imag, type, s, dim, norm, overwrite_x)
        return out

    shape, dim = _init_nd_shape_and_dims(x, s, dim)
    x = _promote_dtype(x)

    if len(dim) == 0:
        return x

    for n, dim1 in zip(shape, dim):
        x = idct(
            x, type=type, n=n, dim=dim1, norm=norm, overwrite_x=overwrite_x
        )
    return x


def dstn(x, type=2, s=None, dim=None, norm=None, overwrite_x=False):
    """Compute a multidimensional Discrete Sine Transform.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DST (see Notes). Default type is 2.
    s : int or array_like of ints or None, optional
        The shape of the result. If both `s` and `dim` (see below) are None,
        `s` is ``x.shape``; if `s` is None but `dim` is not None, then `s` is
        ``numpy.take(x.shape, dim, dim=0)``.
        If ``s[i] > x.shape[i]``, the ith dimension is padded with zeros.
        If ``s[i] < x.shape[i]``, the ith dimension is truncated to length
        ``s[i]``.
        If any element of `s` is -1, the size of the corresponding dimension of
        `x` is used.
    dim : int or array_like of ints or None, optional
        Axes over which the DST is computed. If not given, the last ``len(s)``
        dimensions are used, or all dimensions if `s` is also not specified.
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    y : torch.Tensor of real
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.dstn`

    Notes
    -----
    For full details of the DST types and normalization modes, as well as
    references, see :func:`scipy.fft.dst`.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = dstn(x.real, type, s, dim, norm, overwrite_x)
        out = out + 1j * dstn(x.imag, type, s, dim, norm, overwrite_x)
        return out

    shape, dim = _init_nd_shape_and_dims(x, s, dim)
    x = _promote_dtype(x)

    if len(dim) == 0:
        return x

    for n, dim1 in zip(shape, dim):
        x = dst(
            x, type=type, n=n, dim=dim1, norm=norm, overwrite_x=overwrite_x
        )
    return x


def idstn(x, type=2, s=None, dim=None, norm=None, overwrite_x=False):
    """Compute a multidimensional Discrete Sine Transform.

    Parameters
    ----------
    x : torch.Tensor
        The input array.
    type : {1, 2, 3, 4}, optional
        Type of the DST (see Notes). Default type is 2.
    s : int or array_like of ints or None, optional
        The shape of the result. If both `s` and `dim` (see below) are None,
        `s` is ``x.shape``; if `s` is None but `dim` is not None, then `s` is
        ``numpy.take(x.shape, dim, dim=0)``.
        If ``s[i] > x.shape[i]``, the ith dimension is padded with zeros.
        If ``s[i] < x.shape[i]``, the ith dimension is truncated to length
        ``s[i]``.
        If any element of `s` is -1, the size of the corresponding dimension of
        `x` is used.
    dim : int or array_like of ints or None, optional
        Axes over which the IDST is computed. If not given, the last ``len(s)``
        dimensions are used, or all dimensions if `s` is also not specified.
    norm : {"backward", "ortho", "forward"}, optional
        Normalization mode (see Notes). Default is "backward".
    overwrite_x : bool, optional
        If True, the contents of `x` can be destroyed; the default is False.

    Returns
    -------
    y : torch.Tensor of real
        The transformed input array.

    See Also
    --------
    :func:`scipy.fft.idstn`

    Notes
    -----
    For full details of the IDST types and normalization modes, as well as
    references, see :func:`scipy.fft.idst`.
    """
    if x.dtype.is_complex:
        # separable application on real and imaginary parts
        out = idstn(x.real, type, s, dim, norm, overwrite_x)
        out = out + 1j * idstn(x.imag, type, s, dim, norm, overwrite_x)
        return out

    shape, dim = _init_nd_shape_and_dims(x, s, dim)
    x = _promote_dtype(x)

    if len(dim) == 0:
        return x

    for n, dim1 in zip(shape, dim):
        x = idst(
            x, type=type, n=n, dim=dim1, norm=norm, overwrite_x=overwrite_x
        )
    return x


def _iterable_of_int(x, name=None):
    """Convert ``x`` to an iterable sequence of int."""
    if isinstance(x, numbers.Number):
        x = (x,)

    try:
        x = [operator.index(a) for a in x]
    except TypeError as e:
        name = name or 'value'
        raise ValueError(
            f'{name} must be a scalar or iterable of integers'
        ) from e

    return x


def _init_nd_shape_and_dims(x, shape, dims):
    """Handles shape and dims arguments for nd transforms."""
    noshape = shape is None
    nodims = dims is None

    if not nodims:
        dims = _iterable_of_int(dims, 'dim')
        dims = [a + x.ndim if a < 0 else a for a in dims]

        if any(a >= x.ndim or a < 0 for a in dims):
            raise ValueError('dims exceeds dimensionality of input')
        if len(set(dims)) != len(dims):
            raise ValueError('all dims must be unique')

    if not noshape:
        shape = _iterable_of_int(shape, 'shape')
        nshape = len(shape)
        if dims and len(dims) != nshape:
            raise ValueError(
                'when given, dims and shape arguments'
                ' have to be of the same length'
            )
        if nodims:
            if nshape > x.ndim:
                raise ValueError('shape requires more dims than are present')
            dims = range(x.ndim - len(shape), x.ndim)

        shape = [x.shape[a] if s == -1 else s for s, a in zip(shape, dims)]
    elif nodims:
        shape = list(x.shape)
        dims = range(x.ndim)
    else:
        shape = [x.shape[a] for a in dims]

    if any(s < 1 for s in shape):
        raise ValueError(
            f'invalid number of data points ({shape}) specified'
        )

    return shape, dims


sqrt2 = math.sqrt(2)


def ortho_prescale_(
        x, dim, type, dst=False, inverse=False, square=False, scipy=False
):
    if not isinstance(dim, numbers.Number):
        for d in dim:
            x = ortho_prescale_(x, d, type, dst, inverse, square, scipy)
        return x

    f = 0.5 * sqrt2 if inverse else sqrt2
    if square:
        f *= f
    x = x.movedim(dim, 0)
    if type == 3:
        index = -1 if (dst and not scipy) else 0
        x[index] *= f
    elif type == 1 and not dst:
        x[0] *= f
        x[-1] *= f
    x = x.movedim(0, dim)
    return x


def ortho_postscale_(
        x, dim, type, dst=False, inverse=False, square=False, scipy=False
):
    if not isinstance(dim, numbers.Number):
        for d in dim:
            x = ortho_postscale_(x, d, type, dst, inverse, square, scipy)
        return x

    f = sqrt2 if inverse else 0.5 * sqrt2
    if square:
        f *= f
    x = x.movedim(dim, 0)
    if type == 2:
        index = -1 if (dst and not scipy) else 0
        x[index] *= f
    elif type == 1 and not dst:
        x[0] *= f
        x[-1] *= f
    x = x.movedim(0, dim)
    return x


def get_ortho_scale(x, dim, type):
    if not isinstance(dim, numbers.Number):
        f = 1
        for d in dim:
            f *= get_ortho_scale(x, d, type)
        return f
    n = x.shape[dim]
    delta = -1 if type == 1 else 0
    f = 1 / math.sqrt(2 * (n + delta))
    return f


def ortho_scale_(x, dim, type, inverse=False, square=False):
    f = get_ortho_scale(x, dim, type)
    if square:
        f *= f
    if inverse:
        f = 1 / f
    x *= f
    return x


def _cook_shape(a, s, dims, value_type):
    if s is None or s == a.shape:
        return a
    if (value_type == 'C2R') and (s[-1] is not None):
        s = list(s)
        s[-1] = s[-1] // 2 + 1
    for sz, dim in zip(s, dims):
        if (sz is not None) and (sz != a.shape[dim]):
            shape = list(a.shape)
            if shape[dim] > sz:
                index = [slice(None)] * a.ndim
                index[dim] = slice(0, sz)
                a = a[tuple(index)]
            else:
                index = [slice(None)] * a.ndim
                index[dim] = slice(0, shape[dim])
                shape[dim] = sz
                z = a.new_zeros(shape)  # C layout by default
                z[tuple(index)] = a
                a = z
    return a


def _promote_dtype(x):
    dtype = x.dtype
    if not dtype.is_floating_point:
        # use float64 instead of promote_types to match SciPy's behavior
        float_dtype = torch.float64
    elif dtype is torch.float16:
        float_dtype = torch.float32
    else:
        float_dtype = dtype
    assert float_dtype in (torch.float32, torch.float64)
    return x.to(float_dtype, copy=False)


def _get_dct_norm_factor(n, inorm, dct_type=2):
    """Normalization factors for DCT/DST I-IV.

    Parameters
    ----------
    n : int
        Data size.
    inorm : {'none', 'sqrt', 'full'}
        When `inorm` is 'none', the scaling factor is 1.0 (unnormalized). When
        `inorm` is 'sqrt', scaling by ``1/sqrt(d)`` as needed for an orthogonal
        transform is used. When `inorm` is 'full', normalization by ``1/d`` is
        applied. The value of ``d`` depends on both `n` and the `dct_type`.
    dct_type : {1, 2, 3, 4}
        Which type of DCT or DST is being normalized?.

    Returns
    -------
    fct : float
        The normalization factor.
    """
    if inorm == 'none':
        return 1
    delta = -1 if dct_type == 1 else 0
    d = 2 * (n + delta)
    if inorm == 'full':
        fct = 1 / d
    elif inorm == 'sqrt':
        fct = 1 / math.sqrt(d)
    else:
        raise ValueError('expected inorm = "none", "sqrt" or "full"')
    return fct


def _reshuffle_dct2(x, dim, dst=False):
    """Reorder entries to allow computation of DCT/DST-II via FFT."""
    sl_even = [slice(None)] * x.ndim
    sl_even[dim] = slice(0, None, 2)
    sl_even = tuple(sl_even)
    sl_odd = [slice(None)] * x.ndim
    sl_odd[dim] = slice(1, None, 2)
    sl_odd = tuple(sl_odd)
    if dst:
        x = torch.cat((x[sl_even], -x[sl_odd].flip(dim)), dim=dim)
    else:
        x = torch.cat((x[sl_even], x[sl_odd].flip(dim)), dim=dim)
    return x


def _mult_factor_dct2(
        n, n_truncate, norm_factor, dtype=torch.float32, device=None
):
    real = torch.zeros(n_truncate, dtype=dtype, device=device)
    imag = torch.arange(n_truncate, dtype=dtype, device=device)
    imag *= -math.pi / (2 * n)
    out = torch.complex(real, imag).exp_().mul_(2 * norm_factor)
    return out


def _exp_factor_dct2(x, n, dim, norm_factor, n_truncate=None):
    """Twiddle & scaling factors for computation of DCT/DST-II via FFT."""
    if n_truncate is None:
        n_truncate = n
    tmp = _mult_factor_dct2(n, n_truncate, norm_factor, x.dtype, x.device)

    if x.ndim == 1:
        return tmp
    tmp_shape = [1] * x.ndim
    tmp_shape[dim] = n_truncate
    tmp_shape = tuple(tmp_shape)
    return tmp.reshape(tmp_shape)


def _dct_or_dst_type2(
    x, n=None, dim=-1, forward=True, norm=None, dst=False,
):
    """Forward DCT/DST-II (or inverse DCT/DST-III) along a single dim

    Parameters
    ----------
    x : torch.Tensor
        The data to transform.
    n : int
        The size of the transform. If None, ``x.shape[dim]`` is used.
    dim : int
        Axis along which the transform is applied.
    forward : bool
        Set true to indicate that this is a forward DCT-II as opposed to an
        inverse DCT-III (The difference between the two is only in the
        normalization factor).
    norm : {None, 'ortho', 'forward', 'backward', 'ortho_scipy'}
        The normalization convention to use.
        If `'ortho_scipy'`, use scipy's incorrect `'ortho`` normalization.
    dst : bool
        If True, a discrete sine transform is computed rather than the discrete
        cosine transform.
    overwrite_x : bool
        Indicates that it is okay to overwrite x. In practice, the current
        implementation never performs the transform in-place.

    Returns
    -------
    y: torch.Tensor
        The transformed array.
    """
    if dim < -x.ndim or dim >= x.ndim:
        raise IndexError('dim out of range')
    if dim < 0:
        dim += x.ndim
    if n is not None and n < 1:
        raise ValueError(
            f'invalid number of data points ({n}) specified'
        )

    x = _cook_shape(x, (n,), (dim,), 'R2R')
    n = x.shape[dim]

    x = _reshuffle_dct2(x, dim, dst)

    ortho_scipy = norm == 'ortho_scipy'
    if ortho_scipy:
        norm = 'ortho'

    if norm == 'ortho':
        inorm = 'sqrt'
    elif norm == 'forward':
        inorm = 'full' if forward else 'none'
    else:
        inorm = 'none' if forward else 'full'
    norm_factor = _get_dct_norm_factor(n, inorm=inorm, dct_type=2)

    x = _fft.fft(x, n=n, dim=dim)
    tmp = _exp_factor_dct2(x.real, n, dim, norm_factor)
    x *= tmp  # broadcasting
    x = torch.real(x)

    if dst and ortho_scipy:
        x = x.flip(dim)

    if norm == 'ortho':
        x = x.movedim(dim, 0)
        x[0] *= sqrt2 * 0.5
        x = x.movedim(0, dim)

    if dst and not ortho_scipy:
        x = x.flip(dim)

    return x


def _reshuffle_dct3(y, n, dim, dst):
    """Reorder entries to allow computation of DCT/DST-II via FFT."""
    x = torch.empty_like(y)
    n_half = (n + 1) // 2

    # Store first half of y in the even entries of the output
    sl_even = [slice(None)] * y.ndim
    sl_even[dim] = slice(0, None, 2)
    sl_even = tuple(sl_even)

    sl_half = [slice(None)] * y.ndim
    sl_half[dim] = slice(0, n_half)
    x[sl_even] = y[tuple(sl_half)]

    # Store the second half of y in the odd entries of the output
    sl_odd = [slice(None)] * y.ndim
    sl_odd[dim] = slice(1, None, 2)
    sl_odd = tuple(sl_odd)

    # sl_half[dim] = slice(-1, n_half - 1, -1)
    sl_half[dim] = slice(n_half, None)
    if dst:
        x[sl_odd] = -y[tuple(sl_half)].flip(dim)
    else:
        x[sl_odd] = y[tuple(sl_half)].flip(dim)
    return x


def _mult_factor_dct3(n, norm_factor, dtype=torch.float32, device=None):
    real = torch.zeros(n, dtype=dtype, device=device)
    imag = torch.arange(n, dtype=dtype, device=device)
    imag *= math.pi / (2 * n)
    out = torch.complex(real, imag).exp_().mul_(2 * norm_factor * n)
    return out


# _mult_factor_dct3 = ElementwiseKernel(
#     in_params='R xr, int32 N, R norm_factor',
#     out_params='C y',
#     operation="""
#     C j(0., 1.);
#     y = (R)(2 * N * norm_factor) * exp(j * (R)(i * M_PI / (2 * N)));""",
# )


def _exp_factor_dct3(x, n, dim, norm_factor):
    """Twiddle & scaling factors for computation of DCT/DST-III via FFT."""
    tmp = _mult_factor_dct3(n, norm_factor, x.dtype, x.device)
    if x.ndim == 1:
        return tmp
    # prepare shape for broadcasting along non-transformed axes
    tmp_shape = [1] * x.ndim
    tmp_shape[dim] = n
    tmp_shape = tuple(tmp_shape)
    return tmp.reshape(tmp_shape)


def _dct_or_dst_type3(
    x, n=None, dim=-1, norm=None, forward=True, dst=False,
):
    """Forward DCT/DST-III (or inverse DCT/DST-II) along a single dim.

    Parameters
    ----------
    x : torch.Tensor
        The data to transform.
    n : int
        The size of the transform. If None, ``x.shape[dim]`` is used.
    dim : int
        Axis along which the transform is applied.
    forward : bool
        Set true to indicate that this is a forward DCT-III as opposed to an
        inverse DCT-II (The difference between the two is only in the
        normalization factor).
    norm : {None, 'ortho', 'forward', 'backward', 'ortho_scipy'}
        The normalization convention to use.
        If `'ortho_scipy'`, use scipy's incorrect `'ortho`` normalization.
    dst : bool
        If True, a discrete sine transform is computed rather than the discrete
        cosine transform.
    overwrite_x : bool
        Indicates that it is okay to overwrite x. In practice, the current
        implementation never performs the transform in-place.

    Returns
    -------
    y: torch.Tensor
        The transformed array.

    """
    if dim < -x.ndim or dim >= x.ndim:
        raise IndexError('dim out of range')
    if dim < 0:
        dim += x.ndim
    if n is not None and n < 1:
        raise ValueError(
            f'invalid number of data points ({n}) specified'
        )

    x = _cook_shape(x, (n,), (dim,), 'R2R')
    n = x.shape[dim]

    ortho_scipy = norm == 'ortho_scipy'
    if ortho_scipy:
        norm = 'ortho'

    # determine normalization factor
    if norm == 'ortho':
        sl0_scale = 0.5 * sqrt2
        inorm = 'sqrt'
    elif norm == 'forward':
        sl0_scale = 0.5
        inorm = 'full' if forward else 'none'
    elif norm == 'backward' or norm is None:
        sl0_scale = 0.5
        inorm = 'none' if forward else 'full'
    else:
        raise ValueError(f'Invalid norm value "{norm}", should be "backward", '
                         '"ortho" or "forward"')
    norm_factor = _get_dct_norm_factor(n, inorm=inorm, dct_type=3)

    if dst:
        x = x.flip(dim)
        if ortho_scipy:
            sl0 = [slice(None)] * x.ndim
            sl0[dim] = slice(-1, None)
            x[tuple(sl0)] *= sqrt2
            sl0_scale = 0.5

    # scale by exponentials and normalization factor
    tmp = _exp_factor_dct3(x, n, dim, norm_factor)
    x = x * tmp  # broadcasting

    sl0 = [slice(None)] * x.ndim
    sl0[dim] = slice(1)
    x[tuple(sl0)] *= sl0_scale

    # inverse fft
    x = _fft.ifft(x, n=n, dim=dim)
    x = torch.real(x)

    # reorder entries
    return _reshuffle_dct3(x, n, dim, dst)


def _dct_type1(
    x, n=None, dim=-1, norm=None, forward=True,
):
    if dim < -x.ndim or dim >= x.ndim:
        raise IndexError('dim out of range')
    if dim < 0:
        dim += x.ndim
    if n is not None and n < 1:
        raise ValueError(
            f'invalid number of data points ({n}) specified'
        )

    x = _cook_shape(x, (n,), (dim,), 'R2R')
    n = x.shape[dim]

    # replicate signal
    slicer_post = [slice(None)] * x.ndim
    slicer_post[dim] = slice(1, -1)
    x_post = x[tuple(slicer_post)]
    x = torch.cat([x, x_post.flip(dim)], dim=dim)

    # orthogonalization step (pre-transform)
    if norm.startswith('ortho'):
        x = x.movedim(dim, 0)
        x[0] *= sqrt2
        x[n-1] *= sqrt2
        x = x.movedim(0, dim)

    # determine normalization factor
    if norm.startswith('ortho'):
        inorm = 'sqrt'
    elif norm == 'forward':
        inorm = 'full' if forward else 'none'
    elif norm == 'backward' or norm is None:
        inorm = 'none' if forward else 'full'
    else:
        raise ValueError(f'Invalid norm value "{norm}", should be "backward", '
                         '"ortho" or "forward"')
    norm_factor = _get_dct_norm_factor(n, inorm=inorm, dct_type=1)

    # fft
    x = _fft.fft(x, n=2*(n-1), dim=dim)
    slicer_pre = [slice(None)] * x.ndim
    slicer_pre[dim] = slice(n)
    x = torch.real(x[tuple(slicer_pre)])
    x *= norm_factor

    # orthogonalization step (post-transform)
    if norm.startswith('ortho'):
        x = x.movedim(dim, 0)
        x[0] /= sqrt2
        x[-1] /= sqrt2
        x = x.movedim(0, dim)

    return x


def _dst_type1(
        x, n=None, dim=-1, norm=None, forward=True,
):
    if dim < -x.ndim or dim >= x.ndim:
        raise IndexError('dim out of range')
    if dim < 0:
        dim += x.ndim
    if n is not None and n < 1:
        raise ValueError(
            f'invalid number of data points ({n}) specified'
        )

    x = _cook_shape(x, (n,), (dim,), 'R2R')
    n = x.shape[dim]

    # replicate signal
    slicer_pre = [slice(None)] * x.ndim
    slicer_pre[dim] = slice(1, n+1)
    slicer_post = [slice(None)] * x.ndim
    slicer_post[dim] = slice(n+2, None)
    bigshape = list(x.shape)
    bigshape[dim] = 2*(n+1)
    tmp = x
    x = x.new_zeros(bigshape)
    x[tuple(slicer_pre)] = tmp
    x[tuple(slicer_post)] = tmp.flip(dim)
    x[tuple(slicer_post)] *= -1
    del tmp

    # determine normalization factor
    if norm.startswith('ortho'):
        inorm = 'sqrt'
    elif norm == 'forward':
        inorm = 'full' if forward else 'none'
    elif norm == 'backward' or norm is None:
        inorm = 'none' if forward else 'full'
    else:
        raise ValueError(f'Invalid norm value "{norm}", should be "backward", '
                         '"ortho" or "forward"')
    norm_factor = _get_dct_norm_factor(n+2, inorm=inorm, dct_type=1)
    # TODO: I am not exactly sure why I need the `+2`

    # fft
    x = _fft.fft(x, n=2*(n+1), dim=dim)
    slicer_pre = [slice(None)] * x.ndim
    slicer_pre[dim] = slice(1, n+1)
    x = torch.imag(x[tuple(slicer_pre)])
    x *= -norm_factor

    return x
